import abc
from typing import Optional

from rosny.state import CommonState
from rosny.utils import setup_logger, default_object_name


class AbstractStream(metaclass=abc.ABCMeta):
    def on_compile_begin(self):
        pass

    def on_compile_end(self):
        pass

    @abc.abstractmethod
    def compile(self,
                common_state: Optional[CommonState] = None,
                name: Optional[str] = None,
                handle_signals: bool = True):
        pass

    def on_start_begin(self):
        pass

    def on_start_end(self):
        pass

    @abc.abstractmethod
    def start(self):
        pass

    @abc.abstractmethod
    def wait(self, timeout: Optional[float] = None):
        pass

    def on_stop_begin(self):
        pass

    def on_stop_end(self):
        pass

    @abc.abstractmethod
    def stop(self):
        pass

    def on_join_begin(self):
        pass

    def on_join_end(self):
        pass

    @abc.abstractmethod
    def join(self, timeout: Optional[float] = None):
        pass

    @abc.abstractmethod
    def stopped(self) -> bool:
        pass

    @abc.abstractmethod
    def compiled(self) -> bool:
        pass

    @abc.abstractmethod
    def joined(self) -> bool:
        pass


class BaseStream(AbstractStream, metaclass=abc.ABCMeta):
    def __init__(self):
        self.name = default_object_name(self)
        self.logger = setup_logger(self.name)
        self.common_state = CommonState()
        self._compiled = False
        self.handle_signals = True

    def compile(self,
                common_state: Optional[CommonState] = None,
                name: Optional[str] = None,
                handle_signals: bool = True):
        self.name = self.__class__.__name__ if name is None else name
        self.logger = setup_logger(self.name)
        if common_state is not None:
            self.common_state = common_state
        self.handle_signals = handle_signals
        self._compiled = True

    def wait(self, timeout: Optional[float] = None):
        self.logger.info(f"Waiting stream with timeout {timeout}")
        self.common_state.wait_exit(timeout=timeout)
        if self.common_state.exit_is_set():
            self.logger.info("Waiting ended, exit event is set")
        else:
            self.logger.info("Waiting ended, timeout exceeded")

    def compiled(self) -> bool:
        return self._compiled

    def __del__(self):
        if not self.stopped():
            self.stop()



import time
from typing import Optional

from rosny.abstract import BaseStream


class LoopTimeMeter:
    def __init__(self):
        self.mean = 0.0
        self.count = 0
        self.last_time = time.perf_counter()

    def reset(self):
        self.mean = 0.0
        self.count = 0
        self.last_time = time.perf_counter()

    def start(self):
        self.last_time = time.perf_counter()

    def end(self):
        self.count += 1
        now_time = time.perf_counter()
        delta = now_time - self.last_time
        self.mean += (delta - self.mean) / self.count
        self.last_time = now_time


class LoopRateManager:
    def __init__(self,
                 loop_rate: Optional[float] = None,
                 min_sleep: float = 1e-9):
        self._loop_rate = None
        self._loop_time = None
        self._sleep_delay = None
        self._time_meter = LoopTimeMeter()
        self._prev_time = time.perf_counter()

        self.loop_rate = loop_rate
        self.min_sleep = min_sleep

    def _build(self, loop_rate):
        self._loop_rate = loop_rate
        if loop_rate is None:
            self._loop_time = None
            self._sleep_delay = None
        else:
            self._loop_time = 1.0 / self.loop_rate
            self._sleep_delay = 0.
        self._time_meter.reset()
        self._prev_time = time.perf_counter()

    def reset(self):
        self._build(self._loop_rate)

    @property
    def loop_rate(self):
        return self._loop_rate

    @loop_rate.setter
    def loop_rate(self, value):
        self._build(value)

    def timing(self):
        if self._loop_rate is None:
            if self.min_sleep:
                time.sleep(self.min_sleep)
        else:
            self._time_meter.end()
            self._sleep_delay += self._time_meter.mean - self._loop_time
            self._sleep_delay = max(self._sleep_delay, 0)

            sleep_time = (self._loop_time
                          + self._prev_time
                          - time.perf_counter()
                          - self._sleep_delay)
            sleep_time = max(self.min_sleep, sleep_time)

            time.sleep(sleep_time)
            self._prev_time = time.perf_counter()


class Profiler:
    def __init__(self, stream: BaseStream, interval: Optional[float] = None):
        self._stream = stream
        self.interval = interval
        self._time_meter = LoopTimeMeter()
        self._last_profile_time = time.perf_counter()

    def reset(self, stream: BaseStream):
        self._stream = stream
        self._time_meter.reset()
        self._last_profile_time = time.perf_counter()

    def profile(self):
        if self.interval is not None:
            self._time_meter.end()
            if self._time_meter.last_time - self._last_profile_time > self.interval:
                loop_time = self._time_meter.mean
                loop_rate = 1 / loop_time if loop_time else float('inf')
                self._stream.common_state.profile_stats[self._stream.name] = loop_time
                self._stream.logger.info(
                    f"Profile - loop time {loop_time:.6g}, loop rate {loop_rate:.4g}"
                )

                self._time_meter.reset()
                self._last_profile_time = time.perf_counter()



import abc
from typing import Optional, Union

from rosny.state import CommonState
from rosny.abstract import BaseStream
from rosny.timing import LoopRateManager, Profiler
from rosny.signal import start_signals, stop_signals


class LoopStream(BaseStream, metaclass=abc.ABCMeta):
    def __init__(self,
                 loop_rate: Optional[float] = None,
                 min_sleep: float = 1e-9,
                 profile_interval: Optional[float] = None,
                 daemon: bool = False):
        super().__init__()
        self.daemon = daemon
        self._driver = None
        self.rate_manager = LoopRateManager(loop_rate=loop_rate,
                                            min_sleep=min_sleep)
        self.profiler = Profiler(stream=self, interval=profile_interval)

    @abc.abstractmethod
    def work(self):
        pass

    def work_loop(self):
        self.on_work_loop_begin()
        try:
            self.rate_manager.reset()
            self.profiler.reset(self)
            while not self.stopped():
                self.work()
                self.rate_manager.timing()
                self.profiler.profile()
        except (Exception, KeyboardInterrupt) as exception:
            self.on_catch_exception(exception)
        finally:
            self.on_work_loop_end()

    def on_catch_exception(self, exception: Union[Exception, KeyboardInterrupt]):
        self.logger.exception(exception)
        self.common_state.set_exit()

    def on_work_loop_begin(self):
        pass

    def on_work_loop_end(self):
        pass

    @abc.abstractmethod
    def _start_driver(self):
        pass

    @abc.abstractmethod
    def _stop_driver(self):
        pass

    @abc.abstractmethod
    def _join_driver(self, timeout):
        pass

    def compile(self,
                common_state: Optional[CommonState] = None,
                name: Optional[str] = None,
                handle_signals: bool = True):
        self.on_compile_begin()
        super().compile(common_state=common_state,
                        name=name,
                        handle_signals=handle_signals)
        self.on_compile_end()

    def start(self):
        self.logger.info("Starting stream")
        if self.stopped():
            if self.joined():
                if not self.compiled():
                    self.compile()
                self.common_state.clear_exit()
                self.on_start_begin()
                self._start_driver()
                if self.handle_signals:
                    start_signals(self)
                self.on_start_end()
                self.logger.info("Stream started")
            else:
                self.logger.error("Stream stopped but not joined")
        else:
            self.logger.error("Stream is already started")

    def stop(self):
        self.logger.info("Stopping stream")
        if not self.stopped():
            self.on_stop_begin()
            self._stop_driver()
            if self.handle_signals:
                stop_signals(self)
            self.on_stop_end()
            self.logger.info("Stream stopped")
        else:
            self.logger.error("Stream is already stopped")

    def join(self, timeout: Optional[float] = None):
        self.logger.info("Joining stream")
        if not self.joined():
            self.on_join_begin()
            self._join_driver(timeout=timeout)
            self.on_join_end()
            self.logger.info("Stream joined")
        else:
            self.logger.error("Stream is already joined")

    def joined(self) -> bool:
        return self._driver is None



import abc
from typing import Optional
from multiprocessing import Process, Value

from rosny.loop import LoopStream
from rosny.utils import setup_logger


class ProcessStream(LoopStream, metaclass=abc.ABCMeta):
    def __init__(self,
                 loop_rate: Optional[float] = None,
                 min_sleep: float = 1e-9,
                 profile_interval: Optional[float] = None,
                 daemon: bool = False):
        super().__init__(loop_rate=loop_rate,
                         min_sleep=min_sleep,
                         profile_interval=profile_interval,
                         daemon=daemon)
        self._driver = None
        self._stopped = Value('i', 1)

    def work_loop(self):
        self.logger = setup_logger(self.name)  # necessary for spawn and forkserver
        super().work_loop()

    def _start_driver(self):
        self._driver = Process(target=self.work_loop,
                               name=self.name,
                               daemon=self.daemon)
        self.logger.info(f"Starting process {self.name}")
        self._stopped.value = 0
        self._driver.start()

    def _stop_driver(self):
        self._stopped.value = 1

    def _join_driver(self, timeout):
        self._driver.join(timeout)
        if self._driver.is_alive():
            self.logger.error(f"Process '{self._driver}' join timeout {timeout}")
        else:
            self._driver = None
            self.common_state.clear_exit()

    def stopped(self) -> bool:
        return self._stopped.value



from typing import Optional
from multiprocessing import Event, Manager


class CommonState:
    def __init__(self):
        self._manager: Optional[Manager] = Manager()
        self.profile_stats: dict = self._manager.dict()
        self._exit_event = Event()

    def set_exit(self):
        self._exit_event.set()

    def clear_exit(self):
        self._exit_event.clear()

    def wait_exit(self, timeout: Optional[float] = None):
        self._exit_event.wait(timeout=timeout)

    def exit_is_set(self) -> bool:
        return self._exit_event.is_set()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["_manager"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._manager = None



import abc
import time
from typing import Optional, Dict

from rosny.abstract import BaseStream, AbstractStream
from rosny.state import CommonState
from rosny.signal import start_signals, stop_signals


class ComposeStream(BaseStream, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __init__(self):
        super().__init__()
        self._streams: Dict[str, AbstractStream] = dict()

    def __setattr__(self, name, value):
        if isinstance(value, AbstractStream):
            self._streams[name] = value
        object.__setattr__(self, name, value)

    def compile(self,
                common_state: Optional[CommonState] = None,
                name: Optional[str] = None,
                handle_signals: bool = True):
        self.on_compile_begin()
        super().compile(common_state=common_state,
                        name=name,
                        handle_signals=handle_signals)
        for stream_name, stream in self._streams.items():
            stream.compile(
                common_state=self.common_state,
                name=f"{self.name}/{stream_name}",
                handle_signals=False
            )
        self.on_compile_end()

    def start(self):
        if not self.compiled():
            self.compile()
        self.logger.info("Starting stream")
        self.on_start_begin()
        for stream in self._streams.values():
            stream.start()
        if self.handle_signals:
            start_signals(self)
        self.on_start_end()
        self.logger.info("Stream started")

    def stop(self):
        self.logger.info("Stopping stream")
        self.on_stop_begin()
        for stream in self._streams.values():
            stream.stop()
        if self.handle_signals:
            stop_signals(self)
        self.on_stop_end()
        self.logger.info("Stream stopped")

    def join(self, timeout: Optional[float] = None):
        self.logger.info("Joining stream")
        self.on_join_begin()
        for stream in self._streams.values():
            start = time.perf_counter()
            stream.join(timeout=timeout)
            if timeout is not None:
                timeout -= time.perf_counter() - start
                timeout = max(timeout, 0)
        self.on_join_end()
        self.logger.info("Stream joined")

    def stopped(self) -> bool:
        for stream in self._streams.values():
            if not stream.stopped():
                return False
        return True

    def joined(self) -> bool:
        for stream in self._streams.values():
            if not stream.joined():
                return False
        return True



import sys
import logging


def default_object_name(object_: object) -> str:
    return f"{object_.__class__.__name__}-{str(id(object_))}"


def setup_logger(name: str) -> logging.Logger:
    formatter = logging.Formatter('[%(asctime)s][%(levelname)s] %(name)s: %(message)s')
    stdout = logging.StreamHandler(stream=sys.stdout)
    stdout.setLevel(logging.INFO)
    stdout.setFormatter(formatter)

    logger = logging.getLogger(name)
    logger.setLevel(logging.INFO)
    if logger.hasHandlers():
        logger.handlers.clear()
    logger.addHandler(stdout)
    return logger



import signal

_signals = [signal.SIGINT, signal.SIGTERM]
_default_handlers = {sig: signal.getsignal(signal.SIGINT) for sig in _signals}


class SignalException(BaseException):
    def __init__(self, signum, frame):
        message = f"Handle signal: {signal.Signals(signum).name}"
        super().__init__(message)
        self.signum = signum
        self.frame = frame


def start_signals(stream):
    def signal_handler(signum, frame):
        exception = SignalException(signum, frame)
        stream.logger.error(exception)
        stream.stop()
        raise exception

    for sig in _signals:
        signal.signal(sig, signal_handler)
    stream.logger.info("Start handling signals")


def stop_signals(stream):
    for sig in _signals:
        signal.signal(sig, _default_handlers[sig])
    stream.logger.info("Stop handling signals")



import abc
from threading import Thread
from typing import Optional

from rosny.loop import LoopStream


class ThreadStream(LoopStream, metaclass=abc.ABCMeta):
    def __init__(self,
                 loop_rate: Optional[float] = None,
                 min_sleep: float = 1e-9,
                 profile_interval: Optional[float] = None,
                 daemon: bool = False):
        super().__init__(loop_rate=loop_rate,
                         min_sleep=min_sleep,
                         profile_interval=profile_interval,
                         daemon=daemon)
        self._driver = None
        self._stopped = True

    def _start_driver(self):
        self._driver = Thread(target=self.work_loop,
                              name=self.name,
                              daemon=self.daemon)
        self.logger.info(f"Starting thread {self.name}")
        self._stopped = False
        self._driver.start()

    def _stop_driver(self):
        self._stopped = True

    def _join_driver(self, timeout):
        self._driver.join(timeout)
        if self._driver.is_alive():
            self.logger.error(f"Thread '{self._driver}' join timeout {timeout}")
        else:
            self._driver = None
            self.common_state.clear_exit()

    def stopped(self) -> bool:
        return self._stopped



import os
import io
import re
from setuptools import setup, find_packages


def read(*names, **kwargs):
    with io.open(os.path.join(os.path.dirname(__file__), *names),
                 encoding=kwargs.get("encoding", "utf8")) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='rosny',
    version=find_version('rosny', '__init__.py'),
    author='Ruslan Baikulov',
    author_email='ruslan1123@gmail.com',
    url='https://github.com/lRomul/rosny',
    description='A simple library for building concurrency applications.',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    license='MIT',
    packages=find_packages(exclude=("tests", "tests.*",)),
    zip_safe=True,
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
